# ActivitySim
# See full license in LICENSE.txt.
import logging
from builtins import range

import numpy as np
import pandas as pd

from activitysim.abm.models.util.trip import (
    cleanup_failed_trips,
    flag_failed_trip_leg_mates,
)
from activitysim.abm.tables.size_terms import tour_destination_size_terms
from activitysim.core import (
    assign,
    chunk,
    config,
    expressions,
    inject,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_sample import interaction_sample
from activitysim.core.interaction_sample_simulate import interaction_sample_simulate
from activitysim.core.pathbuilder import TransitVirtualPathBuilder
from activitysim.core.skim_dictionary import DataFrameMatrix
from activitysim.core.tracing import print_elapsed_time
from activitysim.core.util import assign_in_place, reindex

from .util import estimation

logger = logging.getLogger(__name__)

NO_DESTINATION = -1

# TRIP_ORIG_TAZ = 'TAZ'
ALT_DEST_TAZ = "ALT_DEST_TAZ"
# PRIMARY_DEST_TAZ = 'PRIMARY_DEST_TAZ'
# DEST_MAZ = 'dest_maz'


def _destination_sample(
    primary_purpose,
    trips,
    alternatives,
    model_settings,
    size_term_matrix,
    skims,
    alt_dest_col_name,
    estimator,
    chunk_size,
    chunk_tag,
    trace_label,
):
    """

    Note: trips with no viable destination receive no sample rows
    (because we call interaction_sample with allow_zero_probs=True)
    All other trips will have one or more rows with pick_count summing to sample_size

    returns
        choices: pandas.DataFrame

               alt_dest      prob  pick_count
    trip_id
    102829169      2898  0.002333           1
    102829169      2901  0.004976           1
    102829169      3193  0.002628           1
    """

    spec = simulate.spec_for_segment(
        model_settings,
        spec_id="DESTINATION_SAMPLE_SPEC",
        segment_name=primary_purpose,
        estimator=estimator,
    )

    sample_size = model_settings["SAMPLE_SIZE"]
    if config.setting("disable_destination_sampling", False) or (
        estimator and estimator.want_unsampled_alternatives
    ):
        # FIXME interaction_sample will return unsampled complete alternatives with probs and pick_count
        logger.info(
            "Estimation mode for %s using unsampled alternatives short_circuit_choices"
            % (trace_label,)
        )
        sample_size = 0

    locals_dict = config.get_model_constants(model_settings).copy()

    # size_terms of destination zones are purpose-specific, and trips have various purposes
    # so the relevant size_term for each interaction_sample row
    # cannot be determined until after choosers are joined with alternatives
    # (unless we iterate over trip.purpose - which we could, though we are already iterating over trip_num)
    # so, instead, expressions determine row-specific size_term by a call to: size_terms.get(df.alt_dest, df.purpose)
    locals_dict.update({"size_terms": size_term_matrix})
    locals_dict.update(skims)

    log_alt_losers = config.setting("log_alt_losers", False)

    choices = interaction_sample(
        choosers=trips,
        alternatives=alternatives,
        sample_size=sample_size,
        alt_col_name=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        allow_zero_probs=True,
        spec=spec,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return choices


def destination_sample(
    primary_purpose,
    trips,
    alternatives,
    model_settings,
    size_term_matrix,
    skim_hotel,
    estimator,
    chunk_size,
    trace_label,
):

    chunk_tag = "trip_destination.sample"

    skims = skim_hotel.sample_skims(presample=False)
    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    choices = _destination_sample(
        primary_purpose,
        trips,
        alternatives,
        model_settings,
        size_term_matrix,
        skims,
        alt_dest_col_name,
        estimator,
        chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    return choices


def aggregate_size_term_matrix(maz_size_term_matrix, maz_taz):

    df = maz_size_term_matrix.df
    assert ALT_DEST_TAZ not in df

    dest_taz = df.index.map(maz_taz)
    taz_size_term_matrix = df.groupby(dest_taz).sum()

    taz_size_term_matrix = DataFrameMatrix(taz_size_term_matrix)

    return taz_size_term_matrix


def choose_MAZ_for_TAZ(
    taz_sample, MAZ_size_terms, trips, network_los, alt_dest_col_name, trace_label
):
    """
    Convert taz_sample table with TAZ zone sample choices to a table with a MAZ zone chosen for each TAZ
    choose MAZ probabilistically (proportionally by size_term) from set of MAZ zones in parent TAZ

    Parameters
    ----------
    taz_sample: dataframe with duplicated index <chooser_id_col> and columns: <alt_dest_col_name>, prob, pick_count
    MAZ_size_terms: dataframe with duplicated index <chooser_id_col> and columns: zone_id, dest_TAZ, size_term

    Returns
    -------
    dataframe with with duplicated index <chooser_id_col> and columns: <alt_dest_col_name>, prob, pick_count
    """

    if len(taz_sample) == 0:
        # it can happen that all trips have no viable destinations (and so are dropped from the sample)
        # in which case we can just return the empty taz_sample, since it has the same columns
        return taz_sample.copy()

    # we had to use alt_dest_col_name as specified in model_settings for interaction_sample
    # because expressions reference it to look up size_terms by trip purpose
    DEST_MAZ = alt_dest_col_name
    DEST_TAZ = f"{alt_dest_col_name}_TAZ"

    taz_sample.rename(columns={alt_dest_col_name: DEST_TAZ}, inplace=True)

    trace_hh_id = inject.get_injectable("trace_hh_id", None)
    have_trace_targets = trace_hh_id and tracing.has_trace_targets(taz_sample)
    if have_trace_targets:
        trace_label = tracing.extend_trace_label(trace_label, "choose_MAZ_for_TAZ")

        # write taz choices, pick_counts, probs
        trace_targets = tracing.trace_targets(taz_sample)
        tracing.trace_df(
            taz_sample[trace_targets],
            label=tracing.extend_trace_label(trace_label, "taz_sample"),
            transpose=False,
        )

    # print(f"taz_sample\n{taz_sample}")
    #            alt_dest_TAZ      prob  pick_count
    # trip_id
    # 4343721              12  0.000054           1
    # 4343721              20  0.001864           2

    taz_choices = taz_sample[[DEST_TAZ, "prob"]].reset_index(drop=False)
    taz_choices = taz_choices.reindex(
        taz_choices.index.repeat(taz_sample.pick_count)
    ).reset_index(drop=True)
    taz_choices = taz_choices.rename(columns={"prob": "TAZ_prob"})

    # print(f"taz_choices\n{taz_choices}")
    #         trip_id  alt_dest_TAZ      prob
    # 0       4343721            12  0.000054
    # 1       4343721            20  0.001864
    # 2       4343721            20  0.001864

    # print(f"MAZ_size_terms\n{MAZ_size_terms.df}")
    #           work  escort  shopping  eatout  othmaint  social  othdiscr   univ
    # alt_dest
    # 2         31.0   9.930     0.042   0.258     0.560   0.520    10.856  0.042
    # 3          0.0   3.277     0.029   0.000     0.029   0.029     7.308  0.029
    # 4          0.0   1.879     0.023   0.000     0.023   0.023     5.796  0.023

    # just to make it clear we are siloing choices by chooser_id
    chooser_id_col = (
        taz_sample.index.name
    )  # should be canonical chooser index name (e.g. 'trip_id')

    # for random_for_df, we need df with de-duplicated chooser canonical index
    chooser_df = pd.DataFrame(index=taz_sample.index[~taz_sample.index.duplicated()])
    num_choosers = len(chooser_df)
    assert chooser_df.index.name == chooser_id_col

    # to make choices, <taz_sample_size> rands for each chooser (one rand for each sampled TAZ)
    # taz_sample_size will be model_settings['SAMPLE_SIZE'] samples, except if we are estimating
    taz_sample_size = taz_choices.groupby(chooser_id_col)[DEST_TAZ].count().max()

    # taz_choices index values should be contiguous
    assert (
        taz_choices[chooser_id_col] == np.repeat(chooser_df.index, taz_sample_size)
    ).all()

    # we need to choose a MAZ for each DEST_TAZ choice
    # probability of choosing MAZ based on MAZ size_term fraction of TAZ total
    # there will be a different set (and number) of candidate MAZs for each TAZ
    # (preserve index, which will have duplicates as result of join)

    maz_taz = network_los.maz_taz_df[["MAZ", "TAZ"]].rename(
        columns={"TAZ": DEST_TAZ, "MAZ": DEST_MAZ}
    )
    maz_sizes = pd.merge(
        taz_choices[[chooser_id_col, DEST_TAZ]].reset_index(),
        maz_taz,
        how="left",
        on=DEST_TAZ,
    ).set_index("index")

    purpose = maz_sizes["trip_id"].map(trips.purpose)  # size term varies by purpose
    maz_sizes["size_term"] = MAZ_size_terms.get(maz_sizes[DEST_MAZ], purpose)

    # print(f"maz_sizes\n{maz_sizes}")
    #          trip_id  alt_dest_TAZ  alt_dest  size_term
    # index
    # 0        4343721            12      3445      0.019
    # 0        4343721            12     11583      0.017
    # 0        4343721            12     21142      0.020

    if have_trace_targets:
        # write maz_sizes: maz_sizes[index,trip_id,dest_TAZ,zone_id,size_term]
        maz_sizes_trace_targets = tracing.trace_targets(maz_sizes, slicer="trip_id")
        trace_maz_sizes = maz_sizes[maz_sizes_trace_targets]
        tracing.trace_df(
            trace_maz_sizes,
            label=tracing.extend_trace_label(trace_label, "maz_sizes"),
            transpose=False,
        )

    # number of DEST_TAZ candidates per chooser
    maz_counts = maz_sizes.groupby(maz_sizes.index).size().values
    # print(maz_counts)

    # max number of MAZs for any TAZ
    max_maz_count = maz_counts.max()
    # print(f"max_maz_count {max_maz_count}")

    # offsets of the first and last rows of each chooser in sparse interaction_utilities
    last_row_offsets = maz_counts.cumsum()
    first_row_offsets = np.insert(last_row_offsets[:-1], 0, 0)

    # repeat the row offsets once for each dummy utility to insert
    # (we want to insert dummy utilities at the END of the list of alternative utilities)
    # inserts is a list of the indices at which we want to do the insertions
    inserts = np.repeat(last_row_offsets, max_maz_count - maz_counts)

    # insert zero filler to pad each alternative set to same size
    padded_maz_sizes = np.insert(maz_sizes.size_term.values, inserts, 0.0)
    padded_maz_sizes = padded_maz_sizes.reshape(-1, max_maz_count)

    # prob array with one row TAZ_choice, one column per alternative
    row_sums = padded_maz_sizes.sum(axis=1)
    maz_probs = np.divide(padded_maz_sizes, row_sums.reshape(-1, 1))
    assert maz_probs.shape == (num_choosers * taz_sample_size, max_maz_count)

    rands = (
        pipeline.get_rn_generator()
        .random_for_df(chooser_df, n=taz_sample_size)
        .reshape(-1, 1)
    )
    assert len(rands) == num_choosers * taz_sample_size
    assert len(rands) == maz_probs.shape[0]

    # make choices
    # positions is array with the chosen alternative represented as a column index in probs
    # which is an integer between zero and max_maz_count
    positions = np.argmax((maz_probs.cumsum(axis=1) - rands) > 0.0, axis=1)

    # shouldn't have chosen any of the dummy pad positions
    assert (positions < maz_counts).all()

    taz_choices[DEST_MAZ] = maz_sizes[DEST_MAZ].take(positions + first_row_offsets)
    taz_choices["MAZ_prob"] = maz_probs[np.arange(maz_probs.shape[0]), positions]
    taz_choices["prob"] = taz_choices["TAZ_prob"] * taz_choices["MAZ_prob"]

    if have_trace_targets:

        taz_choices_trace_targets = tracing.trace_targets(taz_choices, slicer="trip_id")
        trace_taz_choices_df = taz_choices[taz_choices_trace_targets]
        tracing.trace_df(
            trace_taz_choices_df,
            label=tracing.extend_trace_label(trace_label, "taz_choices"),
            transpose=False,
        )

        lhs_df = trace_taz_choices_df[["trip_id", DEST_TAZ]]
        alt_dest_columns = [f"dest_maz_{c}" for c in range(max_maz_count)]

        # following the same logic as the full code, but for trace cutout
        trace_maz_counts = maz_counts[taz_choices_trace_targets]
        trace_last_row_offsets = maz_counts[taz_choices_trace_targets].cumsum()
        trace_inserts = np.repeat(
            trace_last_row_offsets, max_maz_count - trace_maz_counts
        )

        # trace dest_maz_alts
        padded_maz_sizes = np.insert(
            trace_maz_sizes[DEST_MAZ].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_alts"),
            transpose=False,
        )

        # trace dest_maz_size_terms
        padded_maz_sizes = np.insert(
            trace_maz_sizes["size_term"].values, trace_inserts, 0.0
        ).reshape(-1, max_maz_count)
        df = pd.DataFrame(
            data=padded_maz_sizes,
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_size_terms"),
            transpose=False,
        )

        # trace dest_maz_probs
        df = pd.DataFrame(
            data=maz_probs[taz_choices_trace_targets],
            columns=alt_dest_columns,
            index=trace_taz_choices_df.index,
        )
        df = pd.concat([lhs_df, df], axis=1)
        df["rand"] = rands[taz_choices_trace_targets]
        tracing.trace_df(
            df,
            label=tracing.extend_trace_label(trace_label, "dest_maz_probs"),
            transpose=False,
        )

    taz_choices = taz_choices.drop(columns=["TAZ_prob", "MAZ_prob"])
    taz_choices = taz_choices.groupby([chooser_id_col, DEST_MAZ]).agg(
        prob=("prob", "max"), pick_count=("prob", "count")
    )

    taz_choices.reset_index(level=DEST_MAZ, inplace=True)

    return taz_choices


def destination_presample(
    primary_purpose,
    trips,
    alternatives,
    model_settings,
    size_term_matrix,
    skim_hotel,
    network_los,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    trace_label = tracing.extend_trace_label(trace_label, "presample")
    chunk_tag = "trip_destination.presample"  # distinguish from trip_destination.sample

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]
    maz_taz = network_los.maz_taz_df[["MAZ", "TAZ"]].set_index("MAZ").TAZ

    TAZ_size_term_matrix = aggregate_size_term_matrix(size_term_matrix, maz_taz)

    TRIP_ORIGIN = model_settings["TRIP_ORIGIN"]
    PRIMARY_DEST = model_settings["PRIMARY_DEST"]
    trips_taz = trips.copy()

    trips_taz[TRIP_ORIGIN] = trips_taz[TRIP_ORIGIN].map(maz_taz)
    trips_taz[PRIMARY_DEST] = trips_taz[PRIMARY_DEST].map(maz_taz)

    # alternatives is just an empty dataframe indexed by maz with index name <alt_dest_col_name>
    # but logically, we are aggregating so lets do it, as there is no particular gain in being clever
    alternatives = alternatives.groupby(alternatives.index.map(maz_taz)).sum()

    # # i did this but after changing alt_dest_col_name to 'trip_dest' it
    # # shouldn't be needed anymore
    # alternatives.index.name = ALT_DEST_TAZ

    skims = skim_hotel.sample_skims(presample=True)

    taz_sample = _destination_sample(
        primary_purpose,
        trips_taz,
        alternatives,
        model_settings,
        TAZ_size_term_matrix,
        skims,
        alt_dest_col_name,
        estimator,
        chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
    )

    # choose a MAZ for each DEST_TAZ choice, choice probability based on MAZ size_term fraction of TAZ total
    maz_sample = choose_MAZ_for_TAZ(
        taz_sample, size_term_matrix, trips, network_los, alt_dest_col_name, trace_label
    )

    assert alt_dest_col_name in maz_sample

    return maz_sample


def trip_destination_sample(
    primary_purpose,
    trips,
    alternatives,
    model_settings,
    size_term_matrix,
    skim_hotel,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):
    """

    Returns
    -------
    destination_sample: pandas.dataframe
        choices_df from interaction_sample with (up to) sample_size alts for each chooser row
        index (non unique) is trip_id from trips (duplicated for each alt)
        and columns dest_zone_id, prob, and pick_count

        dest_zone_id: int
            alt identifier from alternatives[<alt_col_name>]
        prob: float
            the probability of the chosen alternative
        pick_count : int
            number of duplicate picks for chooser, alt
    """
    trace_label = tracing.extend_trace_label(trace_label, "sample")

    assert len(trips) > 0
    assert len(alternatives) > 0

    # by default, enable presampling for multizone systems, unless they disable it in settings file
    network_los = inject.get_injectable("network_los")
    pre_sample_taz = network_los.zone_system != los.ONE_ZONE
    if pre_sample_taz and not config.setting("want_dest_choice_presampling", True):
        pre_sample_taz = False
        logger.info(
            f"Disabled destination zone presampling for {trace_label} "
            f"because 'want_dest_choice_presampling' setting is False"
        )

    if pre_sample_taz:

        logger.info(
            "Running %s trip_destination_presample with %d trips"
            % (trace_label, len(trips))
        )

        choices = destination_presample(
            primary_purpose,
            trips,
            alternatives,
            model_settings,
            size_term_matrix,
            skim_hotel,
            network_los,
            estimator,
            chunk_size,
            trace_hh_id,
            trace_label,
        )

    else:
        choices = destination_sample(
            primary_purpose,
            trips,
            alternatives,
            model_settings,
            size_term_matrix,
            skim_hotel,
            estimator,
            chunk_size,
            trace_label,
        )

    return choices


def compute_ood_logsums(
    choosers,
    logsum_settings,
    nest_spec,
    logsum_spec,
    od_skims,
    locals_dict,
    chunk_size,
    trace_label,
    chunk_tag,
):
    """
    Compute one (of two) out-of-direction logsums for destination alternatives

    Will either be trip_origin -> alt_dest or alt_dest -> primary_dest
    """

    locals_dict.update(od_skims)

    # if preprocessor contains tvpb logsums term, `pathbuilder.get_tvpb_logsum()`
    # will get called before a ChunkSizers class object has been instantiated,
    # causing pathbuilder to throw an error at L815 due to the assert statement
    # in `chunk.chunk_log()` at chunk.py L927. To avoid failing this assertion,
    # the preprocessor must be called from within a "null chunker" as follows:
    with chunk.chunk_log(
        tracing.extend_trace_label(trace_label, "annotate_preprocessor"), base=True
    ):
        expressions.annotate_preprocessors(
            choosers, locals_dict, od_skims, logsum_settings, trace_label
        )

    logsums = simulate.simple_simulate_logsums(
        choosers,
        logsum_spec,
        nest_spec,
        skims=od_skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        trace_label=trace_label,
        chunk_tag=chunk_tag,
    )

    assert logsums.index.equals(choosers.index)

    # FIXME not strictly necessary, but would make trace files more legible?
    # logsums = logsums.replace(-np.inf, -999)

    return logsums


def compute_logsums(
    primary_purpose,
    trips,
    destination_sample,
    tours_merged,
    model_settings,
    skim_hotel,
    chunk_size,
    trace_label,
):
    """
    Calculate mode choice logsums using the same recipe as for trip_mode_choice, but do it twice
    for each alternative since we need out-of-direction logsum
    (i.e . origin to alt_dest, and alt_dest to half-tour destination)

    Returns
    -------
        adds od_logsum and dp_logsum columns to trips (in place)
    """
    trace_label = tracing.extend_trace_label(trace_label, "compute_logsums")
    logger.info("Running %s with %d samples", trace_label, destination_sample.shape[0])

    # chunk usage is uniform so better to combine
    chunk_tag = "trip_destination.compute_logsums"

    # FIXME should pass this in?
    network_los = inject.get_injectable("network_los")

    # - trips_merged - merge trips and tours_merged
    trips_merged = pd.merge(
        trips, tours_merged, left_on="tour_id", right_index=True, how="left"
    )
    assert trips_merged.index.equals(trips.index)

    # - choosers - merge destination_sample and trips_merged
    # re/set index because pandas merge does not preserve left index if it has duplicate values!
    choosers = pd.merge(
        destination_sample,
        trips_merged.reset_index(),
        left_index=True,
        right_on="trip_id",
        how="left",
        suffixes=("", "_r"),
    ).set_index("trip_id")
    assert choosers.index.equals(destination_sample.index)

    logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])
    coefficients = simulate.get_segment_coefficients(logsum_settings, primary_purpose)

    nest_spec = config.get_logit_model_settings(logsum_settings)
    nest_spec = simulate.eval_nest_coefficients(nest_spec, coefficients, trace_label)

    logsum_spec = simulate.read_model_spec(file_name=logsum_settings["SPEC"])
    logsum_spec = simulate.eval_coefficients(logsum_spec, coefficients, estimator=None)

    locals_dict = {}
    locals_dict.update(config.get_model_constants(logsum_settings))

    # coefficients can appear in expressions
    locals_dict.update(coefficients)

    skims = skim_hotel.logsum_skims()
    if network_los.zone_system == los.THREE_ZONE:
        # TVPB constants can appear in expressions
        if logsum_settings.get("use_TVPB_constants", True):
            locals_dict.update(
                network_los.setting("TVPB_SETTINGS.tour_mode_choice.CONSTANTS")
            )

    # - od_logsums
    od_skims = {
        "ORIGIN": model_settings["TRIP_ORIGIN"],
        "DESTINATION": model_settings["ALT_DEST_COL_NAME"],
        "odt_skims": skims["odt_skims"],
        "dot_skims": skims["dot_skims"],
        "od_skims": skims["od_skims"],
    }
    if network_los.zone_system == los.THREE_ZONE:
        od_skims.update(
            {
                "tvpb_logsum_odt": skims["tvpb_logsum_odt"],
                "tvpb_logsum_dot": skims["tvpb_logsum_dot"],
            }
        )
    destination_sample["od_logsum"] = compute_ood_logsums(
        choosers,
        logsum_settings,
        nest_spec,
        logsum_spec,
        od_skims,
        locals_dict,
        chunk_size,
        trace_label=tracing.extend_trace_label(trace_label, "od"),
        chunk_tag=chunk_tag,
    )

    # - dp_logsums
    dp_skims = {
        "ORIGIN": model_settings["ALT_DEST_COL_NAME"],
        "DESTINATION": model_settings["PRIMARY_DEST"],
        "odt_skims": skims["dpt_skims"],
        "dot_skims": skims["pdt_skims"],
        "od_skims": skims["dp_skims"],
    }
    if network_los.zone_system == los.THREE_ZONE:
        dp_skims.update(
            {
                "tvpb_logsum_odt": skims["tvpb_logsum_dpt"],
                "tvpb_logsum_dot": skims["tvpb_logsum_pdt"],
            }
        )

    destination_sample["dp_logsum"] = compute_ood_logsums(
        choosers,
        logsum_settings,
        nest_spec,
        logsum_spec,
        dp_skims,
        locals_dict,
        chunk_size,
        trace_label=tracing.extend_trace_label(trace_label, "dp"),
        chunk_tag=chunk_tag,
    )

    return destination_sample


def trip_destination_simulate(
    primary_purpose,
    trips,
    destination_sample,
    model_settings,
    want_logsums,
    size_term_matrix,
    skim_hotel,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):
    """
    Chose destination from destination_sample (with od_logsum and dp_logsum columns added)


    Returns
    -------
    choices - pandas.Series
        destination alt chosen
    """
    trace_label = tracing.extend_trace_label(trace_label, "trip_dest_simulate")
    chunk_tag = "trip_destination.simulate"

    spec = simulate.spec_for_segment(
        model_settings,
        spec_id="DESTINATION_SPEC",
        segment_name=primary_purpose,
        estimator=estimator,
    )

    if estimator:
        estimator.write_choosers(trips)

    alt_dest_col_name = model_settings["ALT_DEST_COL_NAME"]

    logger.info("Running trip_destination_simulate with %d trips", len(trips))

    skims = skim_hotel.sample_skims(presample=False)

    locals_dict = config.get_model_constants(model_settings).copy()
    locals_dict.update({"size_terms": size_term_matrix})
    locals_dict.update(skims)

    log_alt_losers = config.setting("log_alt_losers", False)
    destinations = interaction_sample_simulate(
        choosers=trips,
        alternatives=destination_sample,
        spec=spec,
        choice_column=alt_dest_col_name,
        log_alt_losers=log_alt_losers,
        want_logsums=want_logsums,
        allow_zero_probs=True,
        zero_prob_choice_val=NO_DESTINATION,
        skims=skims,
        locals_d=locals_dict,
        chunk_size=chunk_size,
        chunk_tag=chunk_tag,
        trace_label=trace_label,
        trace_choice_name="trip_dest",
        estimator=estimator,
    )

    if not want_logsums:
        # for consistency, always return a dataframe with canonical column name
        assert isinstance(destinations, pd.Series)
        destinations = destinations.to_frame("choice")

    if estimator:
        # need to overwrite choices here before any failed choices are suppressed
        estimator.write_choices(destinations.choice)

        destinations.choice = estimator.get_survey_values(
            destinations.choice, "trips", "destination"
        )
        estimator.write_override_choices(destinations.choice)

    # drop any failed zero_prob destinations
    if (destinations.choice == NO_DESTINATION).any():
        # logger.debug("dropping %s failed destinations", (destinations == NO_DESTINATION).sum())
        destinations = destinations[destinations.choice != NO_DESTINATION]

    return destinations


def choose_trip_destination(
    primary_purpose,
    trips,
    alternatives,
    tours_merged,
    model_settings,
    want_logsums,
    want_sample_table,
    size_term_matrix,
    skim_hotel,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
):

    logger.info("choose_trip_destination %s with %d trips", trace_label, trips.shape[0])

    t0 = print_elapsed_time()

    # - trip_destination_sample
    destination_sample = trip_destination_sample(
        primary_purpose=primary_purpose,
        trips=trips,
        alternatives=alternatives,
        model_settings=model_settings,
        size_term_matrix=size_term_matrix,
        skim_hotel=skim_hotel,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    dropped_trips = ~trips.index.isin(destination_sample.index.unique())
    if dropped_trips.any():
        logger.warning(
            "%s trip_destination_sample %s trips "
            "without viable destination alternatives"
            % (trace_label, dropped_trips.sum())
        )
        trips = trips[~dropped_trips]

    t0 = print_elapsed_time("%s.trip_destination_sample" % trace_label, t0)

    if trips.empty:
        return pd.Series(index=trips.index).to_frame("choice"), None

    # - compute logsums
    destination_sample = compute_logsums(
        primary_purpose=primary_purpose,
        trips=trips,
        destination_sample=destination_sample,
        tours_merged=tours_merged,
        model_settings=model_settings,
        skim_hotel=skim_hotel,
        chunk_size=chunk_size,
        trace_label=trace_label,
    )

    t0 = print_elapsed_time("%s.compute_logsums" % trace_label, t0)

    # - trip_destination_simulate
    destinations = trip_destination_simulate(
        primary_purpose=primary_purpose,
        trips=trips,
        destination_sample=destination_sample,
        model_settings=model_settings,
        want_logsums=want_logsums,
        size_term_matrix=size_term_matrix,
        skim_hotel=skim_hotel,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    dropped_trips = ~trips.index.isin(destinations.index)
    if dropped_trips.any():
        logger.warning(
            "%s trip_destination_simulate %s trips "
            "without viable destination alternatives"
            % (trace_label, dropped_trips.sum())
        )

    if want_sample_table:
        # FIXME - sample_table
        destination_sample.set_index(
            model_settings["ALT_DEST_COL_NAME"], append=True, inplace=True
        )
    else:
        destination_sample = None

    t0 = print_elapsed_time("%s.trip_destination_simulate" % trace_label, t0)

    return destinations, destination_sample


class SkimHotel(object):
    def __init__(self, model_settings, network_los, trace_label):

        self.model_settings = model_settings
        self.trace_label = tracing.extend_trace_label(trace_label, "skim_hotel")
        self.network_los = network_los
        self.zone_system = network_los.zone_system

    def sample_skims(self, presample):

        o = self.model_settings["TRIP_ORIGIN"]
        d = self.model_settings["ALT_DEST_COL_NAME"]
        p = self.model_settings["PRIMARY_DEST"]

        if presample:
            assert not (self.zone_system == los.ONE_ZONE)
            skim_dict = self.network_los.get_skim_dict("taz")
        else:
            skim_dict = self.network_los.get_default_skim_dict()

        skims = {
            "od_skims": skim_dict.wrap(o, d),
            "dp_skims": skim_dict.wrap(d, p),
            "odt_skims": skim_dict.wrap_3d(
                orig_key=o, dest_key=d, dim3_key="trip_period"
            ),
            "dot_skims": skim_dict.wrap_3d(
                orig_key=d, dest_key=o, dim3_key="trip_period"
            ),
            "dpt_skims": skim_dict.wrap_3d(
                orig_key=d, dest_key=p, dim3_key="trip_period"
            ),
            "pdt_skims": skim_dict.wrap_3d(
                orig_key=p, dest_key=d, dim3_key="trip_period"
            ),
        }

        return skims

    def logsum_skims(self):

        o = self.model_settings["TRIP_ORIGIN"]
        d = self.model_settings["ALT_DEST_COL_NAME"]
        p = self.model_settings["PRIMARY_DEST"]
        skim_dict = self.network_los.get_default_skim_dict()

        skims = {
            "odt_skims": skim_dict.wrap_3d(
                orig_key=o, dest_key=d, dim3_key="trip_period"
            ),
            "dot_skims": skim_dict.wrap_3d(
                orig_key=d, dest_key=o, dim3_key="trip_period"
            ),
            "dpt_skims": skim_dict.wrap_3d(
                orig_key=d, dest_key=p, dim3_key="trip_period"
            ),
            "pdt_skims": skim_dict.wrap_3d(
                orig_key=p, dest_key=d, dim3_key="trip_period"
            ),
            "od_skims": skim_dict.wrap(o, d),
            "dp_skims": skim_dict.wrap(d, p),
        }

        if self.zone_system == los.THREE_ZONE:
            # fixme - is this a lightweight object?
            tvpb = self.network_los.tvpb

            tvpb_logsum_odt = tvpb.wrap_logsum(
                orig_key=o,
                dest_key=d,
                tod_key="trip_period",
                segment_key="demographic_segment",
                trace_label=self.trace_label,
                tag="tvpb_logsum_odt",
            )
            tvpb_logsum_dot = tvpb.wrap_logsum(
                orig_key=d,
                dest_key=o,
                tod_key="trip_period",
                segment_key="demographic_segment",
                trace_label=self.trace_label,
                tag="tvpb_logsum_dot",
            )
            tvpb_logsum_dpt = tvpb.wrap_logsum(
                orig_key=d,
                dest_key=p,
                tod_key="trip_period",
                segment_key="demographic_segment",
                trace_label=self.trace_label,
                tag="tvpb_logsum_dpt",
            )
            tvpb_logsum_pdt = tvpb.wrap_logsum(
                orig_key=p,
                dest_key=d,
                tod_key="trip_period",
                segment_key="demographic_segment",
                trace_label=self.trace_label,
                tag="tvpb_logsum_pdt",
            )

            skims.update(
                {
                    "tvpb_logsum_odt": tvpb_logsum_odt,
                    "tvpb_logsum_dot": tvpb_logsum_dot,
                    "tvpb_logsum_dpt": tvpb_logsum_dpt,
                    "tvpb_logsum_pdt": tvpb_logsum_pdt,
                }
            )

        return skims


def run_trip_destination(
    trips,
    tours_merged,
    estimator,
    chunk_size,
    trace_hh_id,
    trace_label,
    fail_some_trips_for_testing=False,
):
    """
    trip destination - main functionality separated from model step so it can be called iteratively

    Run the trip_destination model, assigning destinations for each (intermediate) trip
    (last trips already have a destination - either the tour primary destination or Home)

    Set trip destination and origin columns, and a boolean failed flag for any failed trips
    (destination for flagged failed trips will be set to -1)

    Parameters
    ----------
    trips
    tours_merged
    want_sample_table
    chunk_size
    trace_hh_id
    trace_label

    Returns
    -------

    """

    model_settings_file_name = "trip_destination.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)
    preprocessor_settings = model_settings.get("preprocessor", None)
    logsum_settings = config.read_model_settings(model_settings["LOGSUM_SETTINGS"])

    logsum_column_name = model_settings.get("DEST_CHOICE_LOGSUM_COLUMN_NAME")
    want_logsums = logsum_column_name is not None

    sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
    want_sample_table = (
        config.setting("want_dest_choice_sample_tables")
        and sample_table_name is not None
    )

    land_use = inject.get_table("land_use")
    size_terms = inject.get_injectable("size_terms")
    network_los = inject.get_injectable("network_los")
    trips = trips.sort_index()
    trips["next_trip_id"] = np.roll(trips.index, -1)
    trips.next_trip_id = trips.next_trip_id.where(trips.trip_num < trips.trip_count, 0)

    # - initialize trip origin and destination to those of half-tour
    # (we will sequentially adjust intermediate trips origin and destination as we choose them)
    # this is now probably redundant with stop_frequency.py L174
    tour_destination = reindex(tours_merged.destination, trips.tour_id).astype(np.int64)
    tour_origin = reindex(tours_merged.origin, trips.tour_id).astype(np.int64)

    # these values are now automatically created when trips are instantiated when
    # stop_frequency step calls trip.initialize_from_tours. But if this module is being
    # called from trip_destination_and_purpose, these columns will have been deleted
    # so they must be re-created
    if pipeline.get_rn_generator().step_name == "trip_purpose_and_destination":
        trips["destination"] = np.where(trips.outbound, tour_destination, tour_origin)
        trips["origin"] = np.where(trips.outbound, tour_origin, tour_destination)
        trips["failed"] = False

    if estimator:
        # need to check or override non-intermediate trip destination
        # should check consistency of survey trips origin, destination with parent tour and subsequent/prior trip?
        # FIXME if not consistent, do we fail or override? (seems weird to override them to bad values?)

        # expect all the same trips
        survey_trips = estimator.get_survey_table("trips").sort_index()
        assert survey_trips.index.equals(trips.index)

        first = survey_trips.trip_num == 1
        last = survey_trips.trip_num == trips.trip_count

        # expect survey's outbound first trip origin to be same as half tour origin
        assert (
            survey_trips.origin[survey_trips.outbound & first]
            == tour_origin[survey_trips.outbound & first]
        ).all()
        # expect outbound last trip destination to be same as half tour destination
        assert (
            survey_trips.destination[survey_trips.outbound & last]
            == tour_destination[survey_trips.outbound & last]
        ).all()

        # expect inbound first trip origin to be same as half tour destination
        assert (
            survey_trips.origin[~survey_trips.outbound & first]
            == tour_destination[~survey_trips.outbound & first]
        ).all()
        # expect inbound last trip destination to be same as half tour origin
        assert (
            survey_trips.destination[~survey_trips.outbound & last]
            == tour_origin[~survey_trips.outbound & last]
        ).all()

    # - filter tours_merged (AFTER copying destination and origin columns to trips)
    # tours_merged is used for logsums, we filter it here upfront to save space and time
    tours_merged_cols = logsum_settings["TOURS_MERGED_CHOOSER_COLUMNS"]
    redundant_cols = model_settings.get("REDUNDANT_TOURS_MERGED_CHOOSER_COLUMNS", [])
    if redundant_cols:
        tours_merged_cols = [c for c in tours_merged_cols if c not in redundant_cols]

    assert model_settings["PRIMARY_DEST"] not in tours_merged_cols
    tours_merged = tours_merged[tours_merged_cols]

    # - skims
    skim_hotel = SkimHotel(model_settings, network_los, trace_label)

    # - size_terms and alternatives
    alternatives = tour_destination_size_terms(land_use, size_terms, "trip")

    # DataFrameMatrix alows us to treat dataframe as virtual a 2-D array, indexed by zone_id, purpose
    # e.g. size_terms.get(df.dest_zone_id, df.purpose)
    # returns a series of size_terms for each chooser's dest_zone_id and purpose with chooser index
    size_term_matrix = DataFrameMatrix(alternatives)

    # don't need size terms in alternatives, just zone_id index
    alternatives = alternatives.drop(alternatives.columns, axis=1)
    alternatives.index.name = model_settings["ALT_DEST_COL_NAME"]

    sample_list = []

    # - process intermediate trips in ascending trip_num order
    intermediate = trips.trip_num < trips.trip_count
    if intermediate.any():

        first_trip_num = trips[intermediate].trip_num.min()
        last_trip_num = trips[intermediate].trip_num.max()

        # iterate over trips in ascending trip_num order
        for trip_num in range(first_trip_num, last_trip_num + 1):

            nth_trips = trips[intermediate & (trips.trip_num == trip_num)]
            nth_trace_label = tracing.extend_trace_label(
                trace_label, "trip_num_%s" % trip_num
            )

            locals_dict = {"network_los": network_los}
            locals_dict.update(config.get_model_constants(model_settings))

            # - annotate nth_trips
            if preprocessor_settings:
                expressions.assign_columns(
                    df=nth_trips,
                    model_settings=preprocessor_settings,
                    locals_dict=locals_dict,
                    trace_label=nth_trace_label,
                )

            logger.info("Running %s with %d trips", nth_trace_label, nth_trips.shape[0])

            # - choose destination for nth_trips, segmented by primary_purpose
            choices_list = []
            for primary_purpose, trips_segment in nth_trips.groupby("primary_purpose"):
                choices, destination_sample = choose_trip_destination(
                    primary_purpose,
                    trips_segment,
                    alternatives,
                    tours_merged,
                    model_settings,
                    want_logsums,
                    want_sample_table,
                    size_term_matrix,
                    skim_hotel,
                    estimator,
                    chunk_size,
                    trace_hh_id,
                    trace_label=tracing.extend_trace_label(
                        nth_trace_label, primary_purpose
                    ),
                )

                choices_list.append(choices)
                if want_sample_table:
                    assert destination_sample is not None
                    sample_list.append(destination_sample)

            destinations_df = pd.concat(choices_list)

            if fail_some_trips_for_testing:
                if len(destinations_df) > 0:
                    destinations_df = destinations_df.drop(destinations_df.index[0])

            failed_trip_ids = nth_trips.index.difference(destinations_df.index)
            if failed_trip_ids.any():
                logger.warning(
                    "%s sidelining %s trips without viable destination alternatives"
                    % (nth_trace_label, failed_trip_ids.shape[0])
                )
                next_trip_ids = nth_trips.next_trip_id.reindex(failed_trip_ids)
                trips.loc[failed_trip_ids, "failed"] = True
                trips.loc[failed_trip_ids, "destination"] = -1
                trips.loc[next_trip_ids, "origin"] = trips.loc[
                    failed_trip_ids
                ].origin.values

            if len(destinations_df) == 0:
                assert failed_trip_ids.all()
                logger.warning(
                    f"all {len(nth_trips)} {primary_purpose} trip_num {trip_num} trips failed"
                )

            if len(destinations_df) > 0:
                # - assign choices to this trip's destinations
                # if estimator, then the choices will already have been overridden by trip_destination_simulate
                # because we need to overwrite choices before any failed choices are suppressed
                assign_in_place(trips, destinations_df.choice.to_frame("destination"))
                if want_logsums:
                    assert "logsum" in destinations_df.columns
                    assign_in_place(
                        trips, destinations_df.logsum.to_frame(logsum_column_name)
                    )

                # - assign choice to next trip's origin
                destinations_df.index = nth_trips.next_trip_id.reindex(
                    destinations_df.index
                )
                assign_in_place(trips, destinations_df.choice.to_frame("origin"))

    del trips["next_trip_id"]

    if len(sample_list) > 0:
        save_sample_df = pd.concat(sample_list)
    else:
        # this could happen if no intermediate trips, or if no saved sample desired
        save_sample_df = None

    return trips, save_sample_df


@inject.step()
def trip_destination(trips, tours_merged, chunk_size, trace_hh_id):
    """
    Choose a destination for all 'intermediate' trips based on trip purpose.

    Final trips already have a destination (the primary tour destination for outbound trips,
    and home for inbound trips.)


    """
    trace_label = "trip_destination"

    model_settings_file_name = "trip_destination.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    CLEANUP = model_settings.get("CLEANUP", True)
    fail_some_trips_for_testing = model_settings.get(
        "fail_some_trips_for_testing", False
    )

    trips_df = trips.to_frame()
    tours_merged_df = tours_merged.to_frame()

    estimator = estimation.manager.begin_estimation("trip_destination")

    if estimator:
        estimator.write_coefficients(model_settings=model_settings)
        # estimator.write_spec(model_settings, tag='SAMPLE_SPEC')
        estimator.write_spec(model_settings, tag="SPEC")
        estimator.set_alt_id(model_settings["ALT_DEST_COL_NAME"])
        estimator.write_table(
            inject.get_injectable("size_terms"), "size_terms", append=False
        )
        estimator.write_table(
            inject.get_table("land_use").to_frame(), "landuse", append=False
        )
        estimator.write_model_settings(model_settings, model_settings_file_name)

    logger.info("Running %s with %d trips", trace_label, trips_df.shape[0])

    trips_df, save_sample_df = run_trip_destination(
        trips_df,
        tours_merged_df,
        estimator=estimator,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
        fail_some_trips_for_testing=fail_some_trips_for_testing,
    )

    # testing feature t0 make sure at least one trip fails so trip_purpose_and_destination model is run
    if (
        config.setting("testing_fail_trip_destination", False)
        and not trips_df.failed.any()
    ):
        if (trips_df.trip_num < trips_df.trip_count).sum() == 0:
            raise RuntimeError(
                f"can't honor 'testing_fail_trip_destination' setting because no intermediate trips"
            )

        fail_o = trips_df[trips_df.trip_num < trips_df.trip_count].origin.max()
        trips_df.failed = (trips_df.origin == fail_o) & (
            trips_df.trip_num < trips_df.trip_count
        )

    if trips_df.failed.any():
        logger.warning("%s %s failed trips", trace_label, trips_df.failed.sum())
        if inject.get_injectable("pipeline_file_prefix", None):
            file_name = f"{trace_label}_failed_trips_{inject.get_injectable('pipeline_file_prefix')}"
        else:
            file_name = f"{trace_label}_failed_trips"
        logger.info("writing failed trips to %s", file_name)
        tracing.write_csv(
            trips_df[trips_df.failed], file_name=file_name, transpose=False
        )

    if estimator:
        estimator.end_estimation()
        # no trips should have failed since we overwrite choices and sample should have not failed trips
        assert not trips_df.failed.any()

    if CLEANUP:

        if trips_df.failed.any():
            flag_failed_trip_leg_mates(trips_df, "failed")

            if save_sample_df is not None:
                save_sample_df.drop(
                    trips_df.index[trips_df.failed], level="trip_id", inplace=True
                )

            trips_df = cleanup_failed_trips(trips_df)

        trips_df.drop(columns="failed", inplace=True, errors="ignore")

    pipeline.replace_table("trips", trips_df)

    if trace_hh_id:
        tracing.trace_df(
            trips_df,
            label=trace_label,
            slicer="trip_id",
            index_label="trip_id",
            warn_if_empty=True,
        )

    if save_sample_df is not None:
        # might be none if want_sample_table but there are no intermediate trips
        # expect samples only for intermediate trip destinations

        assert len(save_sample_df.index.get_level_values(0).unique()) == len(
            trips_df[trips_df.trip_num < trips_df.trip_count]
        )

        sample_table_name = model_settings.get("DEST_CHOICE_SAMPLE_TABLE_NAME")
        assert sample_table_name is not None

        logger.info(
            "adding %s samples to %s" % (len(save_sample_df), sample_table_name)
        )

        # lest they try to put tour samples into the same table
        if pipeline.is_table(sample_table_name):
            raise RuntimeError("sample table %s already exists" % sample_table_name)
        pipeline.extend_table(sample_table_name, save_sample_df)



import argparse
import os

import numpy as np
import openmatrix as omx
import pandas as pd

MAZ_OFFSET = 0

segments = {
    "test": (492, 1100),  # includes univ
    "full": (0, 100000),
}

parser = argparse.ArgumentParser(description="crop SANDAG 2 zone raw_data")
parser.add_argument(
    "segment_name",
    metavar="segment_name",
    type=str,
    nargs=1,
    help=f"geography segmentation (e.g. full)",
)

parser.add_argument(
    "-c",
    "--check_geography",
    default=False,
    action="store_true",
    help="check consistency of MAZ, TAZ zone_ids and foreign keys & write orphan_households file",
)

args = parser.parse_args()


segment_name = args.segment_name[0]
check_geography = args.check_geography

assert segment_name in segments.keys(), f"Unknown seg: {segment_name}"
maz_min, maz_max = segments[segment_name]

input_dir = "./data_raw"
output_dir = f"./data_{segment_name}_2"


print(f"segment_name {segment_name}")

print(f"input_dir {input_dir}")
print(f"output_dir {output_dir}")
print(f"maz_min {maz_min}")
print(f"maz_max {maz_max}")

print(f"check_geography {check_geography}")

if not os.path.isdir(output_dir):
    print(f"creating output directory {output_dir}")
    os.mkdir(output_dir)


def input_path(file_name):
    return os.path.join(input_dir, file_name)


def output_path(file_name):
    return os.path.join(output_dir, file_name)


def integerize_id_columns(df, table_name):
    columns = ["MAZ", "OMAZ", "DMAZ", "TAZ", "zone_id", "household_id", "HHID"]
    for c in df.columns:
        if c in columns:
            print(f"converting {table_name}.{c} to int")
            if df[c].isnull().any():
                print(df[c][df[c].isnull()])
            df[c] = df[c].astype(int)


def read_csv(file_name, integerize=True):
    df = pd.read_csv(input_path(file_name))

    print(f"read {file_name} {df.shape}")

    return df


def to_csv(df, file_name):
    print(f"writing {file_name} {df.shape} {output_path(file_name)}")
    df.to_csv(output_path(file_name), index=False)


print(f"output_dir {output_dir} maz_min {maz_min} maz_max {maz_max}")


if check_geography:

    # ######## check for orphan_households not in any maz in land_use
    land_use = read_csv("land_use.csv")
    land_use = land_use[["MAZ", "TAZ"]]  # King County
    land_use = land_use.sort_values(["TAZ", "MAZ"])

    households = read_csv("households.csv")
    orphan_households = households[~households.MAZ.isin(land_use.MAZ)]
    print(f"{len(orphan_households)} orphan_households")

    # write orphan_households to INPUT directory (since it doesn't belong in output)
    file_name = "orphan_households.csv"
    print(f"writing {file_name} {orphan_households.shape} to {input_path(file_name)}")
    orphan_households.to_csv(input_path(file_name), index=False)

    # ######## check that land_use and maz and taz tables have same MAZs and TAZs

    # could just build maz and taz files, but want to make sure PSRC data is right

    land_use = read_csv("land_use.csv")
    land_use = land_use.sort_values("MAZ")
    maz = read_csv("maz.csv").sort_values("MAZ")

    # ### FATAL ###
    if not land_use.MAZ.isin(maz.MAZ).all():
        print(
            f"land_use.MAZ not in maz.MAZ\n{land_use.MAZ[~land_use.MAZ.isin(maz.MAZ)]}"
        )
        raise RuntimeError(f"land_use.MAZ not in maz.MAZ")

    if not maz.MAZ.isin(land_use.MAZ).all():
        print(f"maz.MAZ not in land_use.MAZ\n{maz.MAZ[~maz.MAZ.isin(land_use.MAZ)]}")

    # ### FATAL ###
    if not land_use.TAZ.isin(maz.TAZ).all():
        print(
            f"land_use.TAZ not in maz.TAZ\n{land_use.TAZ[~land_use.TAZ.isin(maz.TAZ)]}"
        )
        raise RuntimeError(f"land_use.TAZ not in maz.TAZ")

    if not maz.TAZ.isin(land_use.TAZ).all():
        print(f"maz.TAZ not in land_use.TAZ\n{maz.TAZ[~maz.TAZ.isin(land_use.TAZ)]}")

    land_use = land_use.sort_values("TAZ")
    taz = read_csv("taz.csv").sort_values("TAZ")

    # ### FATAL ###
    if not land_use.TAZ.isin(taz.TAZ).all():
        print(
            f"land_use.TAZ not in taz.TAZ\n{land_use.TAZ[~land_use.TAZ.isin(taz.MAZ)]}"
        )
        raise RuntimeError(f"land_use.TAZ not in taz.TAZ")

    if not taz.TAZ.isin(land_use.TAZ).all():
        print(f"taz.TAZ not in land_use.TAZ\n{taz.TAZ[~taz.TAZ.isin(land_use.TAZ)]}")

    # #########s

#
# land_use
#
land_use = read_csv("land_use.csv")
land_use = land_use[(land_use["MAZ"] >= maz_min) & (land_use["MAZ"] <= maz_max)]
integerize_id_columns(land_use, "land_use")
land_use = land_use.sort_values("MAZ")

# make sure we have some HSENROLL and COLLFTE, even for very for small samples
if land_use["HSENROLL"].sum() == 0:
    assert segment_name != "full", f"land_use['HSENROLL'] is 0 for full sample!"
    land_use["HSENROLL"] = land_use["AGE0519"]
    print(f"\nWARNING: land_use.HSENROLL is 0, so backfilled with AGE0519\n")

if land_use["COLLFTE"].sum() == 0:
    assert segment_name != "full", f"land_use['COLLFTE'] is 0 for full sample!"
    land_use["COLLFTE"] = land_use["HSENROLL"]
    print(f"\nWARNING: land_use.COLLFTE is 0, so backfilled with HSENROLL\n")

# move MAZ and TAZ columns to front
land_use = land_use[
    ["MAZ", "TAZ"] + [c for c in land_use.columns if c not in ["MAZ", "TAZ"]]
]
to_csv(land_use, "land_use.csv")

#
# maz
#
maz = read_csv("maz.csv").sort_values(["MAZ", "TAZ"])
maz = maz[maz["MAZ"].isin(land_use.MAZ)]
integerize_id_columns(maz, "maz")

assert land_use.MAZ.isin(maz.MAZ).all()
assert land_use.TAZ.isin(maz.TAZ).all()
assert maz.TAZ.isin(land_use.TAZ).all()
to_csv(maz, "maz.csv")

#
# taz
#
taz = read_csv("taz.csv").sort_values(["TAZ"])
taz = taz[taz["TAZ"].isin(land_use.TAZ)]
integerize_id_columns(taz, "taz")

assert land_use.TAZ.isin(taz.TAZ).all()
to_csv(taz, "taz.csv")

# print(maz.shape)
# print(f"MAZ {len(maz.MAZ.unique())}")
# print(f"TAZ {len(maz.TAZ.unique())}")

#
# households
#
households = read_csv("households.csv")
households = households[households["MAZ"].isin(maz.MAZ)]
integerize_id_columns(households, "households")

to_csv(households, "households.csv")

#
# persons
#
persons = read_csv("persons.csv")
persons = persons[persons["household_id"].isin(households.HHID)]
integerize_id_columns(persons, "persons")

to_csv(persons, "persons.csv")

#
# maz_to_maz_walk and maz_to_maz_bike
#
for file_name in ["maz_to_maz_walk.csv", "maz_to_maz_bike.csv"]:
    m2m = read_csv(file_name)
    m2m = m2m[m2m.OMAZ.isin(maz.MAZ) & m2m.DMAZ.isin(maz.MAZ)]
    integerize_id_columns(m2m, file_name)
    to_csv(m2m, file_name)

#
# skims
#
omx_infile_name = "skims.omx"
skim_data_type = np.float32

omx_in = omx.open_file(input_path(omx_infile_name))
print(f"omx_in shape {omx_in.shape()}")

taz = taz.sort_values("TAZ")
taz.index = taz.TAZ - 1
tazs_indexes = taz.index.tolist()  # index of TAZ in skim (zero-based, no mapping)
taz_labels = taz.TAZ.tolist()  # TAZ zone_ids in omx index order

# create
num_outfiles = 6 if segment_name == "full" else 1
if num_outfiles == 1:
    omx_out = [omx.open_file(output_path(f"skims1.omx"), "w")]
else:
    omx_out = [
        omx.open_file(output_path(f"skims{i+1}.omx"), "w") for i in range(num_outfiles)
    ]

for omx_file in omx_out:
    omx_file.create_mapping("ZONE", taz_labels)

iskim = 0
for mat_name in omx_in.list_matrices():

    # make sure we have a vanilla numpy array, not a CArray
    m = np.asanyarray(omx_in[mat_name]).astype(skim_data_type)
    m = m[tazs_indexes, :][:, tazs_indexes]
    print(f"{mat_name} {m.shape}")

    omx_file = omx_out[iskim % num_outfiles]
    omx_file[mat_name] = m
    iskim += 1


omx_in.close()
for omx_file in omx_out:
    omx_file.close()



# ActivitySim
# See full license in LICENSE.txt.
import os
import subprocess

import pandas as pd
import pandas.testing as pdt
import pkg_resources
import pytest

from activitysim.core import inject


def teardown_function(func):
    inject.clear_cache()
    inject.reinject_decorated_tables()


def example_path(dirname):
    resource = os.path.join("examples", "placeholder_sandag", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def mtc_example_path(dirname):
    resource = os.path.join("examples", "prototype_mtc", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def psrc_example_path(dirname):
    resource = os.path.join("examples", "placeholder_psrc", dirname)
    return pkg_resources.resource_filename("activitysim", resource)


def build_data():
    pass


@pytest.fixture(scope="module")
def data():
    build_data()


def run_test(zone, multiprocess=False):
    def test_path(dirname):
        return os.path.join(os.path.dirname(__file__), dirname)

    def regress(zone):

        # ## regress tours
        regress_tours_df = pd.read_csv(
            test_path(f"regress/final_{zone}_zone_tours.csv")
        )
        tours_df = pd.read_csv(test_path(f"output/final_{zone}_zone_tours.csv"))
        tours_df.to_csv(
            test_path(f"regress/final_{zone}_zone_tours_last_run.csv"), index=False
        )
        print(f"regress tours")
        pdt.assert_frame_equal(tours_df, regress_tours_df, rtol=1e-03)

        # ## regress trips
        regress_trips_df = pd.read_csv(
            test_path(f"regress/final_{zone}_zone_trips.csv")
        )
        trips_df = pd.read_csv(test_path(f"output/final_{zone}_zone_trips.csv"))
        trips_df.to_csv(
            test_path(f"regress/final_{zone}_zone_trips_last_run.csv"), index=False
        )
        print(f"regress trips")
        pdt.assert_frame_equal(trips_df, regress_trips_df, rtol=1e-03)

    # run test
    file_path = os.path.join(os.path.dirname(__file__), "simulation.py")

    if zone == "2":
        base_configs = psrc_example_path(f"configs")
    else:
        base_configs = mtc_example_path(f"configs")

    run_args = [
        "-c",
        test_path(f"configs_{zone}_zone"),
        "-c",
        example_path(f"configs_{zone}_zone"),
        "-c",
        base_configs,
        "-d",
        example_path(f"data_{zone}"),
        "-o",
        test_path("output"),
    ]

    if multiprocess:
        run_args = run_args + ["-s", "settings_mp.yaml"]

    subprocess.run(["coverage", "run", "-a", file_path] + run_args, check=True)

    regress(zone)


def test_1_zone(data):
    run_test(zone="1", multiprocess=False)


def test_1_zone_mp(data):
    run_test(zone="1", multiprocess=True)


def test_2_zone(data):
    run_test(zone="2", multiprocess=False)


def test_2_zone_mp(data):
    run_test(zone="2", multiprocess=True)


def test_3_zone(data):
    run_test(zone="3", multiprocess=False)


def test_3_zone_mp(data):
    run_test(zone="3", multiprocess=True)


if __name__ == "__main__":

    # call each test explicitly so we get a pass/fail for each
    build_data()
    run_test(zone="1", multiprocess=False)
    run_test(zone="1", multiprocess=True)

    run_test(zone="2", multiprocess=False)
    run_test(zone="2", multiprocess=True)

    run_test(zone="3", multiprocess=False)
    run_test(zone="3", multiprocess=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import numpy as np
import pandas as pd

from activitysim.core import config, inject, pipeline

logger = logging.getLogger(__name__)


@inject.step()
def reassign_tour_purpose_by_poe(tours, chunk_size, trace_hh_id):

    """
    Simulates tour purpose choices after tour origin has been assigned. This
    is useful when the original tour purposes are assigned randomly
    from an aggregate distribution that was not segmented by tour origin.
    """

    trace_label = "reassign_tour_purpose_by_poe"
    probs_df = pd.read_csv(config.config_file_path("tour_purpose_probs_by_poe.csv"))
    probs_df.columns = [
        col if col in ["Purpose", "Description"] else int(col)
        for col in probs_df.columns
    ]

    tours_df = tours.to_frame(columns=["tour_type", "poe_id"])
    tour_types = probs_df[["Purpose", "Description"]].set_index("Purpose")[
        "Description"
    ]

    tours_df["purpose_id"] = None
    for poe, group in tours_df.groupby("poe_id"):
        num_tours = len(group)
        purpose_probs = probs_df[poe]
        purpose_cum_probs = purpose_probs.values.cumsum()
        rands = pipeline.get_rn_generator().random_for_df(group)
        purpose_scaled_probs = np.subtract(purpose_cum_probs, rands)
        purpose = np.argmax((purpose_scaled_probs + 1.0).astype("i4"), axis=1)
        tours_df.loc[group.index, "purpose_id"] = purpose
    tours_df["new_tour_type"] = tours_df["purpose_id"].map(tour_types)

    tours = tours.to_frame()
    tours["tour_type"] = tours_df["new_tour_type"].reindex(tours.index)
    tours["purpose_id"] = tours_df["purpose_id"].reindex(tours.index)
    tours["tour_category"] = "non_mandatory"
    tours.loc[tours["tour_type"].isin(["home", "work"]), "tour_category"] = "mandatory"

    pipeline.replace_table("tours", tours)

    return



import os
from pathlib import Path

import numpy as np
import pandas as pd
import yaml
from larch import DataFrames, Model
from larch.util import Dict

from .general import (
    apply_coefficients,
    construct_nesting_tree,
    dict_of_linear_utility_from_spec,
    remove_apostrophes,
)


def stop_frequency_data(
    edb_directory="output/estimation_data_bundle/{name}/",
    settings_file="{name}_model_settings.yaml",
    chooser_data_file="{name}_values_combined.csv",
    values_index_col="tour_id",
):
    name = "stop_frequency"
    edb_directory = edb_directory.format(name=name)

    settings_file = settings_file.format(name=name)
    with open(os.path.join(edb_directory, settings_file), "r") as yf:
        settings = yaml.load(
            yf,
            Loader=yaml.SafeLoader,
        )

    segments = [i["primary_purpose"] for i in settings["SPEC_SEGMENTS"]]

    master_coef = {}
    prior_segs = []
    coef_map = {seg: {} for seg in segments}

    segment_coef = {}
    for seg_ in settings["SPEC_SEGMENTS"]:
        seg_purpose = seg_["primary_purpose"]
        seg_subdir = Path(os.path.join(edb_directory, seg_purpose))
        segment_coef[seg_["primary_purpose"]] = pd.read_csv(
            seg_subdir / seg_["COEFFICIENTS"],
            index_col="coefficient_name",
        )

    for seg in segments:
        for cname, value in segment_coef[seg].value.items():
            if cname in master_coef:
                if master_coef[cname] == value:
                    coef_map[seg][cname] = cname
                else:
                    for pseg in prior_segs:
                        if master_coef.get(f"{cname}_{pseg}", None) == value:
                            coef_map[seg][cname] = f"{cname}_{pseg}"
                            break
                    else:  # no break
                        master_coef[f"{cname}_{seg}"] = value
                        coef_map[seg][cname] = f"{cname}_{seg}"
            else:
                master_coef[cname] = value
                coef_map[seg][cname] = cname
        prior_segs.append(seg)

    # rewrite revised spec files with common segment_coef names
    for seg in segments:
        seg_subdir = Path(os.path.join(edb_directory, seg))
        with open(seg_subdir / f"stop_frequency_SPEC.csv", "rt") as f:
            spec = f.read()
        for kcoef, v in coef_map[seg].items():
            spec = spec.replace(kcoef, v)
        with open(seg_subdir / f"stop_frequency_SPEC_.csv", "wt") as f:
            f.write(spec)

    master_coef_df = pd.DataFrame(data=master_coef, index=["value"]).T
    master_coef_df.index.name = "coefficient_name"

    seg_coefficients = []
    seg_spec = []
    seg_alt_names = []
    seg_alt_codes = []
    seg_alt_names_to_codes = []
    seg_alt_codes_to_names = []
    seg_chooser_data = []

    for seg in settings["SPEC_SEGMENTS"]:
        seg_purpose = seg["primary_purpose"]
        seg_subdir = Path(os.path.join(edb_directory, seg_purpose))
        coeffs_ = pd.read_csv(
            seg_subdir / seg["COEFFICIENTS"], index_col="coefficient_name"
        )
        coeffs_.index = pd.Index(
            [f"{i}_{seg_purpose}" for i in coeffs_.index], name="coefficient_name"
        )
        seg_coefficients.append(coeffs_)
        spec = pd.read_csv(seg_subdir / "stop_frequency_SPEC_.csv")
        spec = remove_apostrophes(spec, ["Label"])
        # spec.iloc[:, 3:] = spec.iloc[:, 3:].applymap(lambda x: f"{x}_{seg_purpose}" if not pd.isna(x) else x)
        seg_spec.append(spec)

        alt_names = list(spec.columns[3:])
        alt_codes = np.arange(1, len(alt_names) + 1)
        alt_names_to_codes = dict(zip(alt_names, alt_codes))
        alt_codes_to_names = dict(zip(alt_codes, alt_names))

        seg_alt_names.append(alt_names)
        seg_alt_codes.append(alt_codes)
        seg_alt_names_to_codes.append(alt_names_to_codes)
        seg_alt_codes_to_names.append(alt_codes_to_names)

        chooser_data = pd.read_csv(
            seg_subdir / chooser_data_file.format(name=name),
            index_col=values_index_col,
        )
        seg_chooser_data.append(chooser_data)

    return Dict(
        edb_directory=Path(edb_directory),
        settings=settings,
        chooser_data=seg_chooser_data,
        coefficients=master_coef_df,
        spec=seg_spec,
        alt_names=seg_alt_names,
        alt_codes=seg_alt_codes,
        alt_names_to_codes=seg_alt_names_to_codes,
        alt_codes_to_names=seg_alt_codes_to_names,
        segments=segments,
        coefficient_map=coef_map,
        segment_coefficients=segment_coef,
    )


def stop_frequency_model(
    edb_directory="output/estimation_data_bundle/{name}/",
    return_data=False,
):
    data = stop_frequency_data(
        edb_directory=edb_directory,
        values_index_col="tour_id",
    )

    models = []

    for n in range(len(data.spec)):

        coefficients = data.coefficients
        # coef_template = data.coef_template # not used
        spec = data.spec[n]
        chooser_data = data.chooser_data[n]
        settings = data.settings

        alt_names = data.alt_names[n]
        alt_codes = data.alt_codes[n]

        from .general import clean_values

        chooser_data = clean_values(
            chooser_data,
            alt_names_to_codes=data.alt_names_to_codes[n],
            choice_code="override_choice_code",
        )

        if settings.get("LOGIT_TYPE") == "NL":
            tree = construct_nesting_tree(data.alt_names[n], settings["NESTS"])
            m = Model(graph=tree)
        else:
            m = Model()

        m.utility_co = dict_of_linear_utility_from_spec(
            spec,
            "Label",
            dict(zip(alt_names, alt_codes)),
        )

        apply_coefficients(coefficients, m)

        avail = True

        d = DataFrames(
            co=chooser_data,
            av=avail,
            alt_codes=alt_codes,
            alt_names=alt_names,
        )

        m.dataservice = d
        m.choice_co_code = "override_choice_code"
        models.append(m)

    from larch.model.model_group import ModelGroup

    models = ModelGroup(models)

    if return_data:
        return (
            models,
            data,
        )

    return models


def update_segment_coefficients(model, data, result_dir=Path("."), output_file=None):
    for m, segment_name in zip(model, data.segments):
        coefficient_map = data.coefficient_map[segment_name]
        segment_c = []
        master_c = []
        for c_local, c_master in coefficient_map.items():
            if c_master in model.pf.index:
                segment_c.append(c_local)
                master_c.append(c_master)
        coefficients = data.segment_coefficients[segment_name].copy()
        coefficients.loc[segment_c, "value"] = model.pf.loc[
            master_c, "value"
        ].to_numpy()
        if output_file is not None:
            os.makedirs(result_dir, exist_ok=True)
            coefficients.reset_index().to_csv(
                result_dir / output_file.format(segment_name=segment_name),
                index=False,
            )



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import config, expressions, inject, pipeline, simulate, tracing
from activitysim.core.util import reindex

from .util import cdap, estimation

logger = logging.getLogger(__name__)


@inject.step()
def cdap_simulate(persons_merged, persons, households, chunk_size, trace_hh_id):
    """
    CDAP stands for Coordinated Daily Activity Pattern, which is a choice of
    high-level activity pattern for each person, in a coordinated way with other
    members of a person's household.

    Because Python requires vectorization of computation, there are some specialized
    routines in the cdap directory of activitysim for this purpose.  This module
    simply applies those utilities using the simulation framework.
    """

    trace_label = "cdap"
    model_settings = config.read_model_settings("cdap.yaml")
    person_type_map = model_settings.get("PERSON_TYPE_MAP", None)
    assert (
        person_type_map is not None
    ), f"Expected to find PERSON_TYPE_MAP setting in cdap.yaml"
    estimator = estimation.manager.begin_estimation("cdap")

    cdap_indiv_spec = simulate.read_model_spec(
        file_name=model_settings["INDIV_AND_HHSIZE1_SPEC"]
    )

    coefficients_df = simulate.read_model_coefficients(model_settings)
    cdap_indiv_spec = simulate.eval_coefficients(
        cdap_indiv_spec, coefficients_df, estimator
    )

    # Rules and coefficients for generating interaction specs for different household sizes
    interaction_coefficients_file_name = model_settings.get(
        "INTERACTION_COEFFICIENTS", "cdap_interaction_coefficients.csv"
    )
    cdap_interaction_coefficients = pd.read_csv(
        config.config_file_path(interaction_coefficients_file_name), comment="#"
    )

    # replace cdap_interaction_coefficients coefficient labels with numeric values
    # for backward compatibility, use where() to allow hard-coded coefficients and dummy (empty) coefficients_file
    coefficients = cdap_interaction_coefficients.coefficient.map(
        coefficients_df.value.to_dict()
    )
    coefficients = cdap_interaction_coefficients.coefficient.where(
        coefficients.isnull(), coefficients
    )
    coefficients = pd.to_numeric(coefficients, errors="coerce").astype(float)
    if coefficients.isnull().any():
        # show them the offending lines from interaction_coefficients_file
        logger.warning(
            f"bad coefficients in INTERACTION_COEFFICIENTS {interaction_coefficients_file_name}\n"
            f"{cdap_interaction_coefficients[coefficients.isnull()]}"
        )
        assert not coefficients.isnull().any()
    cdap_interaction_coefficients.coefficient = coefficients

    """
    spec to compute/specify the relative proportions of each activity (M, N, H)
    that should be used to choose activities for additional household members not handled by CDAP
    This spec is handled much like an activitysim logit utility spec,
    EXCEPT that the values computed are relative proportions, not utilities
    (i.e. values are not exponentiated before being normalized to probabilities summing to 1.0)
    """
    cdap_fixed_relative_proportions = simulate.read_model_spec(
        file_name=model_settings["FIXED_RELATIVE_PROPORTIONS_SPEC"]
    )

    persons_merged = persons_merged.to_frame()

    # add tour-based chunk_id so we can chunk all trips in tour together
    assert "chunk_id" not in persons_merged.columns
    unique_household_ids = persons_merged.household_id.unique()
    household_chunk_ids = pd.Series(
        range(len(unique_household_ids)), index=unique_household_ids
    )
    persons_merged["chunk_id"] = reindex(
        household_chunk_ids, persons_merged.household_id
    )

    constants = config.get_model_constants(model_settings)

    cdap_interaction_coefficients = cdap.preprocess_interaction_coefficients(
        cdap_interaction_coefficients
    )

    # specs are built just-in-time on demand and cached as injectables
    # prebuilding here allows us to write them to the output directory
    # (also when multiprocessing locutor might not see all household sizes)
    logger.info("Pre-building cdap specs")
    for hhsize in range(2, cdap.MAX_HHSIZE + 1):
        spec = cdap.build_cdap_spec(cdap_interaction_coefficients, hhsize, cache=True)
        if inject.get_injectable("locutor", False):
            spec.to_csv(
                config.output_file_path("cdap_spec_%s.csv" % hhsize), index=True
            )

    if estimator:
        estimator.write_model_settings(model_settings, "cdap.yaml")
        estimator.write_spec(model_settings, tag="INDIV_AND_HHSIZE1_SPEC")
        estimator.write_spec(
            model_settings=model_settings, tag="FIXED_RELATIVE_PROPORTIONS_SPEC"
        )
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_table(
            cdap_interaction_coefficients,
            "interaction_coefficients",
            index=False,
            append=False,
        )
        estimator.write_choosers(persons_merged)
        for hhsize in range(2, cdap.MAX_HHSIZE + 1):
            spec = cdap.get_cached_spec(hhsize)
            estimator.write_table(spec, "spec_%s" % hhsize, append=False)

    logger.info("Running cdap_simulate with %d persons", len(persons_merged.index))

    choices = cdap.run_cdap(
        persons=persons_merged,
        person_type_map=person_type_map,
        cdap_indiv_spec=cdap_indiv_spec,
        cdap_interaction_coefficients=cdap_interaction_coefficients,
        cdap_fixed_relative_proportions=cdap_fixed_relative_proportions,
        locals_d=constants,
        chunk_size=chunk_size,
        trace_hh_id=trace_hh_id,
        trace_label=trace_label,
    )

    if estimator:
        estimator.write_choices(choices)
        choices = estimator.get_survey_values(choices, "persons", "cdap_activity")
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # - assign results to persons table and annotate
    persons = persons.to_frame()

    choices = choices.reindex(persons.index)
    persons["cdap_activity"] = choices

    expressions.assign_columns(
        df=persons,
        model_settings=model_settings.get("annotate_persons"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_persons"),
    )

    pipeline.replace_table("persons", persons)

    # - annotate households table
    households = households.to_frame()
    expressions.assign_columns(
        df=households,
        model_settings=model_settings.get("annotate_households"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_households"),
    )
    pipeline.replace_table("households", households)

    tracing.print_summary("cdap_activity", persons.cdap_activity, value_counts=True)
    logger.info(
        "cdap crosstabs:\n%s"
        % pd.crosstab(persons.ptype, persons.cdap_activity, margins=True)
    )



# ActivitySim
# See full license in LICENSE.txt.

import itertools
import logging
import os

import numpy as np
import pandas as pd

from activitysim.core import (
    assign,
    config,
    expressions,
    inject,
    logit,
    los,
    pipeline,
    simulate,
    tracing,
)
from activitysim.core.interaction_simulate import interaction_simulate
from activitysim.core.util import assign_in_place

from .util import estimation

logger = logging.getLogger(__name__)


def append_probabilistic_vehtype_type_choices(choices, model_settings, trace_label):
    """
    Select a fuel type for the provided body type and age of the vehicle.

    Make probabilistic choices based on the `PROBS_SPEC` file.

    Parameters
    ----------
    choices : pandas.DataFrame
        selection of {body_type}_{age} to append vehicle type to
    probs_spec_file : str
    trace_label : str

    Returns
    -------
    choices : pandas.DataFrame
        table of chosen vehicle types
    """
    probs_spec_file = model_settings.get("PROBS_SPEC", None)
    probs_spec = pd.read_csv(config.config_file_path(probs_spec_file), comment="#")

    fleet_year = model_settings.get("FLEET_YEAR")
    probs_spec["age"] = (1 + fleet_year - probs_spec["vehicle_year"]).astype(int)
    probs_spec["vehicle_type"] = (
        probs_spec[["body_type", "age"]].astype(str).agg("_".join, axis=1)
    )

    # left join vehicles to probs
    choosers = pd.merge(
        choices.reset_index(), probs_spec, on="vehicle_type", how="left", indicator=True
    ).set_index("vehicle_id")

    # checking to make sure all alternatives have probabilities
    missing_alts = choosers.loc[choosers._merge == "left_only", ["vehicle_type"]]
    assert (
        len(missing_alts) == 0
    ), f"missing probabilities for alternatives:\n {missing_alts}"

    # chooser columns here should just be the fuel type probabilities
    non_prob_cols = ["_merge", "vehicle_type", "body_type", "age", "vehicle_year"]
    choosers.drop(columns=non_prob_cols, inplace=True)

    # probs should sum to 1 with residual probs resulting in choice of 'fail'
    chooser_probs = choosers.div(choosers.sum(axis=1), axis=0).fillna(0)
    chooser_probs["fail"] = 1 - chooser_probs.sum(axis=1).clip(0, 1)

    # make probabilistic choices
    prob_choices, rands = logit.make_choices(
        chooser_probs, trace_label=trace_label, trace_choosers=choosers
    )

    # convert alt choice index to vehicle type attribute
    prob_choices = chooser_probs.columns[prob_choices.values].to_series(
        index=prob_choices.index
    )
    failed = prob_choices == chooser_probs.columns.get_loc("fail")
    prob_choices = prob_choices.where(~failed, "NOT CHOSEN")

    # add new attribute to logit choice vehicle types
    choices["vehicle_type"] = choices["vehicle_type"] + "_" + prob_choices

    return choices


def annotate_vehicle_type_choice_households(model_settings, trace_label):
    """
    Add columns to the households table in the pipeline according to spec.

    Parameters
    ----------
    model_settings : dict
    trace_label : str
    """
    households = inject.get_table("households").to_frame()
    expressions.assign_columns(
        df=households,
        model_settings=model_settings.get("annotate_households"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_households"),
    )
    pipeline.replace_table("households", households)


def annotate_vehicle_type_choice_persons(model_settings, trace_label):
    """
    Add columns to the persons table in the pipeline according to spec.

    Parameters
    ----------
    model_settings : dict
    trace_label : str
    """
    persons = inject.get_table("persons").to_frame()
    expressions.assign_columns(
        df=persons,
        model_settings=model_settings.get("annotate_persons"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_persons"),
    )
    pipeline.replace_table("persons", households)


def annotate_vehicle_type_choice_vehicles(model_settings, trace_label):
    """
    Add columns to the vehicles table in the pipeline according to spec.

    Parameters
    ----------
    model_settings : dict
    trace_label : str
    """
    vehicles = inject.get_table("vehicles").to_frame()
    expressions.assign_columns(
        df=vehicles,
        model_settings=model_settings.get("annotate_vehicles"),
        trace_label=tracing.extend_trace_label(trace_label, "annotate_vehicles"),
    )
    pipeline.replace_table("vehicles", vehicles)


def get_combinatorial_vehicle_alternatives(alts_cats_dict):
    """
    Build a pandas dataframe containing columns for each vehicle alternative.

    Rows will correspond to the alternative number and will be 0 except for the
    1 in the column corresponding to that alternative.

    Parameters
    ----------
    alts_cats_dict : dict
    model_settings : dict

    Returns
    -------
    alts_wide : pd.DataFrame in wide format expanded using pandas get_dummies function
    alts_long : pd.DataFrame in long format
    """
    cat_cols = list(alts_cats_dict.keys())  # e.g. fuel type, body type, age
    alts_long = pd.DataFrame(
        list(itertools.product(*alts_cats_dict.values())), columns=alts_cats_dict.keys()
    ).astype(str)
    alts_wide = pd.get_dummies(alts_long)  # rows will sum to num_cats

    alts_wide = pd.concat([alts_wide, alts_long], axis=1)
    return alts_wide, alts_long


def construct_model_alternatives(model_settings, alts_cats_dict, vehicle_type_data):
    """
    Construct the table of vehicle type alternatives.

    Vehicle type data is joined to the alternatives table for use in utility expressions.

    Parameters
    ----------
    model_settings : dict
    alts_cats_dict : dict
        nested dictionary of vehicle body, age, and fuel options
    vehicle_type_data : pandas.DataFrame

    Returns
    -------
    alts_wide : pd.DataFrame
        includes column indicators and data for each alternative
    alts_long : pd.DataFrame
        rows just list the alternatives
    """
    probs_spec_file = model_settings.get("PROBS_SPEC", None)
    if probs_spec_file:
        # do not include alternatives from fuel_type if they are given probabilisticly
        del alts_cats_dict["fuel_type"]
    alts_wide, alts_long = get_combinatorial_vehicle_alternatives(alts_cats_dict)

    # merge vehicle type data to alternatives if data is provided
    if (vehicle_type_data is not None) and (probs_spec_file is None):

        alts_wide = pd.merge(
            alts_wide,
            vehicle_type_data,
            how="left",
            on=["body_type", "fuel_type", "age"],
            indicator=True,
        )

        # checking to make sure all alternatives have data
        missing_alts = alts_wide.loc[
            alts_wide._merge == "left_only", ["body_type", "fuel_type", "age"]
        ]

        assert (
            len(missing_alts) == 0
        ), f"missing vehicle data for alternatives:\n {missing_alts}"
        alts_wide.drop(columns="_merge", inplace=True)

    # converting age to integer to allow interactions in utilities
    alts_wide["age"] = alts_wide["age"].astype(int)

    # store alts in primary configs dir for inspection
    configs_dirs = inject.get_injectable("configs_dir")
    configs_dirs = configs_dirs if isinstance(configs_dirs, list) else [configs_dirs]

    if model_settings.get("WRITE_OUT_ALTS_FILE", False):
        alts_wide.to_csv(
            os.path.join(configs_dirs[0]), "vehicle_type_choice_aternatives.csv"
        )

    return alts_wide, alts_long


def get_vehicle_type_data(model_settings, vehicle_type_data_file):
    """
    Read in the vehicle type data and computes the vehicle age.

    Parameters
    ----------
    model_settings : dict
    vehicle_type_data_file : str
        name of vehicle type data file found in config folder

    Returns
    -------
    vehicle_type_data : pandas.DataFrame
        table of vehicle type data with required body_type, age, and fuel_type columns
    """
    vehicle_type_data = pd.read_csv(
        config.config_file_path(vehicle_type_data_file), comment="#"
    )
    fleet_year = model_settings.get("FLEET_YEAR")

    vehicle_type_data["age"] = (
        1 + fleet_year - vehicle_type_data["vehicle_year"]
    ).astype(str)
    vehicle_type_data["vehicle_type"] = (
        vehicle_type_data[["body_type", "age", "fuel_type"]]
        .astype(str)
        .agg("_".join, axis=1)
    )

    return vehicle_type_data


def iterate_vehicle_type_choice(
    vehicles_merged,
    model_settings,
    model_spec,
    locals_dict,
    estimator,
    chunk_size,
    trace_label,
):
    """
    Select vehicle type for each household vehicle sequentially.

    Iterate through household vehicle numbers and select a vehicle type of
    the form {body_type}_{age}_{fuel_type}. The preprocessor is run for each
    iteration on the entire chooser table, not just the one for the current
    vehicle number.  This allows for computation of terms involving the presence
    of other household vehicles.

    Vehicle type data is read in according to the specification and joined to
    the alternatives. It can optionally be included in the output vehicles table
    by specifying the `COLS_TO_INCLUDE_IN_VEHICLE_TABLE` option in the model yaml.

    Parameters
    ----------
    vehicles_merged : orca.DataFrameWrapper
        vehicle list owned by each household merged with households table
    model_settings : dict
        yaml model settings file as dict
    model_spec : pandas.DataFrame
        omnibus spec file with expressions in index and one column per segment
    locals_dict : dict
        additional variables available when writing expressions
    estimator : Estimator object
    chunk_size : orca.injectable
    trace_label : str

    Returns
    -------
    all_choices : pandas.DataFrame
        single table of selected vehicle types and associated data
    all_choosers : pandas.DataFrame
        single table of chooser data with preprocessor variables included
    """
    # - model settings
    nest_spec = config.get_logit_model_settings(model_settings)
    vehicle_type_data_file = model_settings.get("VEHICLE_TYPE_DATA_FILE", None)
    probs_spec_file = model_settings.get("PROBS_SPEC", None)
    alts_cats_dict = model_settings.get("combinatorial_alts", False)

    # adding vehicle type data to be available to locals_dict regardless of option
    if vehicle_type_data_file:
        vehicle_type_data = get_vehicle_type_data(
            model_settings, vehicle_type_data_file
        )
        locals_dict.update({"vehicle_type_data": vehicle_type_data})

    # - Preparing alternatives
    # create alts on-the-fly as cartesian product of categorical values
    if alts_cats_dict:
        # do not include fuel types as alternatives if probability file is supplied
        alts_wide, alts_long = construct_model_alternatives(
            model_settings, alts_cats_dict, vehicle_type_data
        )

    # - preparing choosers for iterating
    vehicles_merged = vehicles_merged.to_frame()
    vehicles_merged["already_owned_veh"] = ""
    logger.info("Running %s with %d vehicles", trace_label, len(vehicles_merged))
    all_choosers = []
    all_choices = []

    # - Selecting each vehicle in the household sequentially
    # This is necessary to determine and use utility terms that include other
    #   household vehicles
    for veh_num in range(1, vehicles_merged.vehicle_num.max() + 1):
        # - preprocessor
        # running preprocessor on entire vehicle table to enumerate vehicle types
        # already owned by the household
        choosers = vehicles_merged
        preprocessor_settings = model_settings.get("preprocessor", None)
        if preprocessor_settings:
            expressions.assign_columns(
                df=choosers,
                model_settings=preprocessor_settings,
                locals_dict=locals_dict,
                trace_label=trace_label,
            )

        # only make choices for vehicles that have not been selected yet
        choosers = choosers[choosers["vehicle_num"] == veh_num]
        logger.info(
            "Running %s for vehicle number %s with %d vehicles",
            trace_label,
            veh_num,
            len(choosers),
        )

        # if there were so many alts that they had to be created programmatically,
        # by combining categorical variables, then the utility expressions should make
        # use of interaction terms to accommodate alt-specific coefficients and constants
        simulation_type = model_settings.get("SIMULATION_TYPE", "interaction_simulate")
        assert (simulation_type == "interaction_simulate") or (
            simulation_type == "simple_simulate"
        ), "SIMULATION_TYPE needs to be interaction_simulate or simple_simulate"

        log_alt_losers = config.setting("log_alt_losers", False)

        if simulation_type == "interaction_simulate":
            assert (
                alts_cats_dict is not None
            ), "Need to supply combinatorial_alts in yaml"

            choices = interaction_simulate(
                choosers=choosers,
                alternatives=alts_wide,
                spec=model_spec,
                log_alt_losers=log_alt_losers,
                locals_d=locals_dict,
                chunk_size=chunk_size,
                trace_label=trace_label,
                trace_choice_name="vehicle_type",
                estimator=estimator,
            )

        # otherwise, "simple simulation" should suffice, with a model spec that enumerates
        # each alternative as a distinct column in the .csv
        elif simulation_type == "simple_simulate":
            choices = simulate.simple_simulate(
                choosers=choosers,
                spec=model_spec,
                log_alt_losers=log_alt_losers,
                nest_spec=nest_spec,
                locals_d=locals_dict,
                chunk_size=chunk_size,
                trace_label=trace_label,
                trace_choice_name="vehicle_type",
                estimator=estimator,
            )

        if isinstance(choices, pd.Series):
            choices = choices.to_frame("choice")

        choices.rename(columns={"choice": "vehicle_type"}, inplace=True)

        if alts_cats_dict:
            alts = (
                alts_long[alts_long.columns]
                .apply(lambda row: "_".join(row.values.astype(str)), axis=1)
                .values
            )
        else:
            alts = model_spec.columns
        choices["vehicle_type"] = choices["vehicle_type"].map(dict(enumerate(alts)))

        # STEP II: append probabilistic vehicle type attributes
        if probs_spec_file is not None:
            choices = append_probabilistic_vehtype_type_choices(
                choices, model_settings, trace_label
            )

        vehicles_merged.loc[choices.index, "already_owned_veh"] = choices[
            "vehicle_type"
        ]
        all_choices.append(choices)
        all_choosers.append(choosers)

    # adding all vehicle numbers to one dataframe
    all_choices = pd.concat(all_choices)
    all_choosers = pd.concat(all_choosers)

    # appending vehicle type data to the vehicle table
    additional_cols = model_settings.get("COLS_TO_INCLUDE_IN_VEHICLE_TABLE")
    if additional_cols:
        additional_cols.append("vehicle_type")
        all_choices = (
            all_choices.reset_index()
            .merge(vehicle_type_data[additional_cols], how="left", on="vehicle_type")
            .set_index("vehicle_id")
        )

    return all_choices, all_choosers


@inject.step()
def vehicle_type_choice(
    persons, households, vehicles, vehicles_merged, chunk_size, trace_hh_id
):
    """Assign a vehicle type to each vehicle in the `vehicles` table.

    If a "SIMULATION_TYPE" is set to simple_simulate in the
    vehicle_type_choice.yaml config file, then the model specification .csv file
    should contain one column of coefficients for each distinct alternative. This
    format corresponds to ActivitySim's :func:`activitysim.core.simulate.simple_simulate`
    format. Otherwise, this model will construct a table of alternatives, at run time,
    based on all possible combinations of values of the categorical variables enumerated
    as "combinatorial_alts" in the .yaml config. In this case, the model leverages
    ActivitySim's :func:`activitysim.core.interaction_simulate` model design, in which
    the model specification .csv has only one column of coefficients, and the utility
    expressions can turn coefficients on or off based on attributes of either
    the chooser _or_ the alternative.

    As an optional second step, the user may also specify a "PROBS_SPEC" .csv file in
    the main .yaml config, corresponding to a lookup table of additional vehicle
    attributes and probabilities to be sampled and assigned to vehicles after the logit
    choices have been made. The rows of the "PROBS_SPEC" file must include all body type
    and vehicle age choices assigned in the logit model. These additional attributes are
    concatenated with the selected alternative from the logit model to form a single
    vehicle type name to be stored in the `vehicles` table as the vehicle_type column.

    Only one household vehicle is selected at a time to allow for the introduction of
    owned vehicle related attributes. For example, a household may be less likely to
    own a second van if they already own one. The model is run sequentially through
    household vehicle numbers. The preprocessor is run for each iteration on the entire
    vehicles table to allow for computation of terms involving the presence of other
    household vehicles.

    The user may also augment the `households` or `persons` tables with new vehicle
    type-based fields specified via expressions in "annotate_households_vehicle_type.csv"
    and "annotate_persons_vehicle_type.csv", respectively.

    Parameters
    ----------
    persons : orca.DataFrameWrapper
    households : orca.DataFrameWrapper
    vehicles : orca.DataFrameWrapper
    vehicles_merged : orca.DataFrameWrapper
    chunk_size : orca.injectable
    trace_hh_id : orca.injectable
    """
    trace_label = "vehicle_type_choice"
    model_settings_file_name = "vehicle_type_choice.yaml"
    model_settings = config.read_model_settings(model_settings_file_name)

    estimator = estimation.manager.begin_estimation("vehicle_type")

    model_spec_raw = simulate.read_model_spec(file_name=model_settings["SPEC"])
    coefficients_df = simulate.read_model_coefficients(model_settings)
    model_spec = simulate.eval_coefficients(model_spec_raw, coefficients_df, estimator)

    constants = config.get_model_constants(model_settings)

    locals_dict = {}
    locals_dict.update(constants)
    locals_dict.update(coefficients_df)

    choices, choosers = iterate_vehicle_type_choice(
        vehicles_merged,
        model_settings,
        model_spec,
        locals_dict,
        estimator,
        chunk_size,
        trace_label,
    )

    if estimator:
        estimator.write_model_settings(model_settings, model_settings_file_name)
        estimator.write_spec(model_settings)
        estimator.write_coefficients(coefficients_df, model_settings)
        estimator.write_choosers(choosers)

        # FIXME #interaction_simulate_estimation_requires_chooser_id_in_df_column
        #  shuold we do it here or have interaction_simulate do it?
        # chooser index must be duplicated in column or it will be omitted from interaction_dataset
        # estimation requires that chooser_id is either in index or a column of interaction_dataset
        # so it can be reformatted (melted) and indexed by chooser_id and alt_id
        assert choosers.index.name == "vehicle_id"
        assert "vehicle_id" not in choosers.columns
        choosers["vehicle_id"] = choosers.index

        # FIXME set_alt_id - do we need this for interaction_simulate estimation bundle tables?
        estimator.set_alt_id("alt_id")
        estimator.set_chooser_id(choosers.index.name)

        estimator.write_choices(choices)
        choices = estimator.get_survey_values(
            choices, "vehicles", "vehicle_type_choice"
        )
        estimator.write_override_choices(choices)
        estimator.end_estimation()

    # update vehicles table
    # vehicles = pd.merge(vehicles.to_frame(), choices, left_index=True, right_index=True)
    vehicles = pd.concat([vehicles.to_frame(), choices], axis=1)
    pipeline.replace_table("vehicles", vehicles)

    # - annotate tables
    if model_settings.get("annotate_households"):
        annotate_vehicle_type_choice_households(model_settings, trace_label)
    if model_settings.get("annotate_persons"):
        annotate_vehicle_type_choice_persons(model_settings, trace_label)
    if model_settings.get("annotate_vehicles"):
        annotate_vehicle_type_choice_vehicles(model_settings, trace_label)

    tracing.print_summary(
        "vehicle_type_choice", vehicles.vehicle_type, value_counts=True
    )

    if trace_hh_id:
        tracing.trace_df(vehicles, label="vehicle_type_choice", warn_if_empty=True)



# ActivitySim
# See full license in LICENSE.txt.
import logging

import pandas as pd

from activitysim.core import inject
from activitysim.core.input import read_input_table

logger = logging.getLogger(__name__)


@inject.table()
def land_use():

    df = read_input_table("land_use")

    # try to make life easy for everybody by keeping everything in canonical order
    # but as long as coalesce_pipeline doesn't sort tables it coalesces, it might not stay in order
    # so even though we do this, anyone downstream who depends on it, should look out for themselves...
    if not df.index.is_monotonic_increasing:
        df = df.sort_index()

    logger.info("loaded land_use %s" % (df.shape,))

    # replace table function with dataframe
    inject.add_table("land_use", df)

    return df


inject.broadcast("land_use", "households", cast_index=True, onto_on="home_zone_id")



import logging
import os

import pandas as pd
import pytest

# import models is necessary to initalize the model steps with orca
from activitysim.abm import models
from activitysim.core import config, pipeline


# Used by conftest.py initialize_pipeline method
@pytest.fixture(scope="module")
def module() -> str:
    """
    A pytest fixture that returns the data folder location.
    :return: folder location for any necessary data to initialize the tests
    """
    return "summarize"


# Used by conftest.py initialize_pipeline method
@pytest.fixture(scope="module")
def tables() -> dict[str, str]:
    """
    A pytest fixture that returns the "mock" tables to build pipeline dataframes. The
    key-value pair is the name of the table and the index column.
    :return: dict
    """
    return {
        "land_use": "zone_id",
        "tours": "tour_id",
        "trips": "trip_id",
        "persons": "person_id",
        "households": "household_id",
    }


# Used by conftest.py initialize_pipeline method
# Set to true if you need to read skims into the pipeline
@pytest.fixture(scope="module")
def initialize_network_los() -> bool:
    """
    A pytest boolean fixture indicating whether network skims should be read from the
    fixtures test data folder.
    :return: bool
    """
    return True


def test_summarize(initialize_pipeline: pipeline.Pipeline, caplog):
    # Run summarize model
    caplog.set_level(logging.DEBUG)
    pipeline.run(models=["summarize"])

    # Retrieve output tables to check contents
    model_settings = config.read_model_settings("summarize.yaml")
    output_location = (
        model_settings["OUTPUT"] if "OUTPUT" in model_settings else "summaries"
    )
    output_dir = config.output_file_path(output_location)

    # Check that households are counted correctly
    households_count = pd.read_csv(
        config.output_file_path(os.path.join(output_location, f"households_count.csv"))
    )
    households = pd.read_csv(config.data_file_path("households.csv"))
    assert int(households_count.iloc[0]) == len(households)

    # Check that bike trips are counted correctly
    trips_by_mode_count = pd.read_csv(
        config.output_file_path(
            os.path.join(output_location, f"trips_by_mode_count.csv")
        )
    )
    trips = pd.read_csv(config.data_file_path("trips.csv"))
    assert int(trips_by_mode_count.BIKE.iloc[0]) == len(
        trips[trips.trip_mode == "BIKE"]
    )



import os
import subprocess
import tempfile

import pandas as pd
import pytest

from activitysim.cli.create import get_example


@pytest.fixture(scope="module")
def est_data():

    cwd = os.getcwd()
    tempdir = tempfile.TemporaryDirectory()
    os.chdir(tempdir.name)

    get_example("example_estimation_sf", "_test_est")
    os.chdir("_test_est")

    # !activitysim run -c configs_estimation/configs -c configs -o output -d data_sf
    print(f"List of files now in {os.getcwd()}")
    subprocess.run(["find", "."])
    print(f"\n\nrunning activitysim estimation mode in {os.getcwd()}")
    subprocess.run(
        [
            "activitysim",
            "run",
            "-c",
            "configs_estimation/configs",
            "-c",
            "configs",
            "-o",
            "output",
            "-d",
            "data_sf",
        ],
    )

    yield os.getcwd()

    os.chdir(cwd)


def _regression_check(dataframe_regression, df, basename=None):
    dataframe_regression.check(
        df.select_dtypes("number")
        .drop(columns=["holdfast"], errors="ignore")
        .clip(-9e9, 9e9),
        # pandas 1.3 handles int8 dtypes as actual numbers, so holdfast needs to be dropped manually
        # we're dropping it not adding to the regression check so older pandas will also work.
        basename=basename,
        default_tolerance=dict(atol=1e-6, rtol=0.1)
        # set a little loose, as there is sometimes a little variance in these
        # results when switching backend implementations. We're checking all
        # the parameters and the log likelihood, so modest variance in individual
        # parameters, especially those with high std errors, is not problematic.
    )


@pytest.mark.parametrize(
    "name,method",
    [
        ("free_parking", "BHHH"),
        ("mandatory_tour_frequency", "SLSQP"),
        ("joint_tour_frequency", "SLSQP"),
        ("joint_tour_composition", "SLSQP"),
        ("joint_tour_participation", "SLSQP"),
        ("mandatory_tour_frequency", "BHHH"),
        ("atwork_subtour_frequency", "SLSQP"),
        ("auto_ownership", "BHHH"),
        ("trip_mode_choice", "SLSQP"),
    ],
)
def test_simple_simulate(est_data, num_regression, dataframe_regression, name, method):
    from activitysim.estimation.larch import component_model

    m = component_model(name)
    m.load_data()
    m.doctor(repair_ch_av="-")
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method=method, options={"maxiter": 1000})
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename=f"test_simple_simulate_{name}_{method}_loglike",
        default_tolerance=dict(atol=1e-6, rtol=1e-3),
    )
    _regression_check(dataframe_regression, m.pf)


@pytest.mark.parametrize(
    "name,method",
    [
        ("workplace_location", "SLSQP"),
        ("school_location", "SLSQP"),
        ("non_mandatory_tour_destination", "SLSQP"),
        ("atwork_subtour_destination", "BHHH"),
        ("trip_destination", "SLSQP"),
    ],
)
def test_location_model(est_data, num_regression, dataframe_regression, name, method):
    from activitysim.estimation.larch import component_model, update_size_spec

    m, data = component_model(name, return_data=True)
    m.load_data()
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method=method, options={"maxiter": 1000})
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename=f"test_loc_{name}_loglike",
    )
    _regression_check(dataframe_regression, m.pf)
    size_spec = update_size_spec(
        m,
        data,
        result_dir=None,
        output_file=None,
    )
    dataframe_regression.check(
        size_spec,
        basename=f"test_loc_{name}_size_spec",
        default_tolerance=dict(atol=1e-6, rtol=5e-2)
        # set a little loose, as there is sometimes a little variance in these
        # results when switching backend implementations.
    )


@pytest.mark.parametrize(
    "name,method",
    [
        ("non_mandatory_tour_scheduling", "SLSQP"),
        ("joint_tour_scheduling", "SLSQP"),
        ("atwork_subtour_scheduling", "SLSQP"),
        ("mandatory_tour_scheduling_work", "SLSQP"),
        ("mandatory_tour_scheduling_school", "SLSQP"),
    ],
)
def test_scheduling_model(est_data, num_regression, dataframe_regression, name, method):
    from activitysim.estimation.larch import component_model, update_size_spec

    m, data = component_model(name, return_data=True)
    m.load_data()
    m.doctor(repair_ch_av="-")
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method=method)
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename=f"test_{name}_loglike",
    )
    _regression_check(dataframe_regression, m.pf)


def test_stop_freq_model(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch import component_model

    name = "stop_frequency"
    m, data = component_model(name, return_data=True)
    m.load_data()
    loglike_prior = m.loglike()
    r = m.maximize_loglike()
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename=f"test_{name}_loglike",
    )
    _regression_check(dataframe_regression, m.pf)


def test_workplace_location(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch import component_model, update_size_spec

    m, data = component_model("workplace_location", return_data=True)
    m.load_data()
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method="SLSQP")
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename="test_workplace_location_loglike",
    )
    _regression_check(dataframe_regression, m.pf)
    size_spec = update_size_spec(
        m,
        data,
        result_dir=None,
        output_file=None,
    )
    dataframe_regression.check(
        size_spec,
        basename="test_workplace_location_size_spec",
        default_tolerance=dict(atol=1e-6, rtol=1e-2),
    )


def test_school_location(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch import component_model, update_size_spec

    m, data = component_model("school_location", return_data=True)
    m.load_data()
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method="BHHH")
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename="test_school_location_loglike",
    )
    _regression_check(dataframe_regression, m.pf)
    size_spec = update_size_spec(
        m,
        data,
        result_dir=None,
        output_file=None,
    )
    dataframe_regression.check(
        size_spec,
        basename="test_school_location_size_spec",
        default_tolerance=dict(atol=1e-6, rtol=1e-2),
    )


def test_cdap_model(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch.cdap import cdap_model

    m = cdap_model()
    m.load_data()
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method="SLSQP", options={"maxiter": 1000})
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename="test_cdap_model_loglike",
    )
    _regression_check(dataframe_regression, m.pf)


def test_nonmand_and_joint_tour_dest_choice(
    est_data, num_regression, dataframe_regression
):
    from activitysim.estimation.larch import component_model

    modelname = ("non_mandatory_tour_destination", "joint_tour_destination")
    m, d = component_model(modelname, return_data=True)
    m.load_data()
    m.doctor(repair_ch_av="-")
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method="SLSQP", options={"maxiter": 1000})
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename="test_nonmand_and_joint_tour_dest_choice_loglike",
    )
    _regression_check(dataframe_regression, m.pf)


def test_tour_and_subtour_mode_choice(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch.mode_choice import (
        atwork_subtour_mode_choice_model,
        tour_mode_choice_model,
    )

    m = tour_mode_choice_model()
    s = atwork_subtour_mode_choice_model()
    m.extend(s)  # join the atwork subtour model to the master group
    m.load_data()
    m.doctor(repair_ch_av="-")
    loglike_prior = m.loglike()
    r = m.maximize_loglike(method="SLSQP", options={"maxiter": 1000})
    num_regression.check(
        {"loglike_prior": loglike_prior, "loglike_converge": r.loglike},
        basename="test_tour_mode_choice_loglike",
    )
    _regression_check(dataframe_regression, m.pf)


def test_nonmand_tour_freq(est_data, num_regression, dataframe_regression):
    from activitysim.estimation.larch.nonmand_tour_freq import nonmand_tour_freq_model

    m = nonmand_tour_freq_model()
    loglike_prior = {}
    for segment_name in m:
        m[segment_name].load_data()
        m[segment_name].doctor(repair_ch_av="-")
        loglike_prior[segment_name] = m[segment_name].loglike()
    r = {}
    for segment_name in m:
        r[segment_name] = m[segment_name].maximize_loglike(
            method="SLSQP", options={"maxiter": 1000}
        )
    loglike_priors = [value for key, value in sorted(loglike_prior.items())]
    loglike_converge = [value.loglike for key, value in sorted(r.items())]
    num_regression.check(
        {"loglike_prior": loglike_priors, "loglike_converge": loglike_converge},
        basename="test_nonmand_tour_freq_loglike",
    )
    _regression_check(dataframe_regression, pd.concat([x.pf for x in m.values()]))
